#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2012,2024  Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2025       Steve Youngs <steve@youngs.cc>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""
Package providing filtering framework for Gramps.
"""

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
import logging
import time

# ------------------------------------------------------------------------
#
# Gramps imports
#
# ------------------------------------------------------------------------
from ..lib.person import Person
from ..lib.family import Family
from ..lib.src import Source
from ..lib.citation import Citation
from ..lib.event import Event
from ..lib.place import Place
from ..lib.repo import Repository
from ..lib.media import Media
from ..lib.note import Note
from ..lib.tag import Tag
from ..const import GRAMPS_LOCALE as glocale
from .rules import Rule
from .optimizer import Optimizer

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import cast, Dict, List, Literal, Set, Tuple
from ..db import Database
from ..types import PrimaryObjectHandle

_ = glocale.translation.gettext
LOG = logging.getLogger(".filter.results")


# -------------------------------------------------------------------------
#
# GenericFilter
#
# -------------------------------------------------------------------------
class GenericFilter:
    """Filter class that consists of zero, one or more rules."""

    logical_functions = ["and", "or", "one"]
    logical_op: Literal["and", "or", "one"]

    def __init__(self, source=None):
        if source:
            self.need_param = source.need_param
            self.flist = source.flist[:]
            self.name = source.name
            self.comment = source.comment
            self.logical_op = source.logical_op
            self.invert = source.invert
        else:
            self.need_param = 0
            self.flist = []
            self.name = ""
            self.comment = ""
            self.logical_op = "and"
            self.invert = False

    def match(self, handle, db):
        """
        Return True or False depending on whether the handle matches the filter.
        """
        obj = self.get_object(db, handle)
        for rule in self.flist:
            rule.requestprepare(db, user=None)

        results = self.apply_to_one(db, obj)

        for rule in self.flist:
            rule.requestreset()

        return results

    def is_empty(self):
        return (len(self.flist) == 0) or (
            len(self.flist) == 1 and ((self.flist[0].is_empty() and not self.invert))
        )

    def set_logical_op(self, val):
        if val in GenericFilter.logical_functions:
            self.logical_op = val
        else:
            raise Exception("invalid operator: %r" % val)

    def set_invert(self, val):
        self.invert = bool(val)

    def get_invert(self):
        return self.invert

    def get_name(self, ulocale=glocale):
        return self.name

    def set_name(self, name):
        self.name = name

    def set_comment(self, comment):
        self.comment = comment

    def get_comment(self):
        return self.comment

    def add_rule(self, rule):
        self.flist.append(rule)

    def delete_rule(self, rule):
        self.flist.remove(rule)

    def set_rules(self, rules):
        self.flist = rules

    def get_rules(self):
        return self.flist

    def get_all_handles(self, db):
        return db.get_person_handles()

    def get_cursor(self, db):
        return db.get_person_cursor()

    def get_tree_cursor(self, db):
        return db.get_person_cursor()

    def make_obj(self):
        return Person()

    def find_from_handle(self, db, handle):
        return db.get_person_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_people()

    def apply_logical_op_to_all(
        self,
        db,
        possible_handles: Set[PrimaryObjectHandle],
        apply_logical_op,
        user=None,
    ):
        LOG.debug(
            "Starting possible_handles: %s",
            len(possible_handles),
        )

        # use the Optimizer to refine the set of possible_handles
        optimizer = Optimizer(self)
        handles_in, handles_out = optimizer.compute_potential_handles_for_filter(self)

        # LOG.debug(
        #    "Optimizer possible_handles: %s",
        #    len(possible_handles),
        # )

        # if user:
        #    user.begin_progress(_("Filter"), _("Applying ..."), len(possible_handles))

        # test each value in possible_handles to compute the final_list
        final_list = []
        for handle in possible_handles:
            if handles_in is not None and handle not in handles_in:
                continue
            if handles_out is not None and handle in handles_out:
                continue

            if user:
                user.step_progress()

            obj = self.get_object(db, handle)

            if apply_logical_op(db, obj, self.flist) != self.invert:
                final_list.append(obj.handle)

        if user:
            user.end_progress()

        return final_list

    def and_test(self, db, data: dict, flist):
        return all(rule.apply_to_one(db, data) for rule in flist)

    def one_test(self, db, data: dict, flist):
        found_one = False
        for rule in flist:
            if rule.apply_to_one(db, data):
                if found_one:
                    return False  # There can be only one!
                found_one = True
        return found_one

    def or_test(self, db, data: dict, flist):
        return any(rule.apply_to_one(db, data) for rule in flist)

    def get_logical_op(self):
        return self.logical_op

    def apply_to_one(self, db, data: dict) -> bool:
        """
        Filter-level apply rules to single data item.
        """
        if self.logical_op == "and":
            res = self.and_test(db, data, self.flist)
        elif self.logical_op == "or":
            res = self.or_test(db, data, self.flist)
        elif self.logical_op == "one":
            res = self.one_test(db, data, self.flist)
        else:
            raise Exception("invalid operator: %r" % self.logical_op)
        return res != self.invert

    def apply(
        self,
        db,
        id_list: List[PrimaryObjectHandle | Tuple] | None = None,
        tupleind: int | None = None,
        user=None,
        tree: bool = False,
    ) -> List[PrimaryObjectHandle | Tuple]:
        """
        Apply the filter using db.

        If id_list and tupleind given, id_list is a list of tuples. tupleind is the
        index of the object handle. So handle_0 = id_list[0][tupleind]. The input
        order in maintained in the output. If multiple tuples have the same handle,
        it is undefined which tuple is returned.

        If id_list given and tupleind is None, id_list if the list of handles to use

        If id_list is None and tree is False, all handles in the database are searched.
        The order of handles in the result is not guaranteed

        If id_list is None and tree is True, all handles in the database are searched.
        The order of handles in the result matches the order of traversal
        by get_tree_cursor

        id_list takes precendence over tree

        user is optional. If present it must be an instance of a User class.

        :Returns: if id_list given, it is returned with the items that
                do not match the filter, filtered out.
                if id_list not given, all items in the database that
                match the filter are returned as a list of handles
        """
        start_time = time.time()
        for rule in self.flist:
            rule.requestprepare(db, user)
        LOG.debug("Prepare time: %s seconds", time.time() - start_time)

        if self.logical_op == "and":
            apply_logical_op = self.and_test
        elif self.logical_op == "or":
            apply_logical_op = self.or_test
        elif self.logical_op == "one":
            apply_logical_op = self.one_test
        else:
            raise Exception("invalid operator: %r" % self.logical_op)

        start_time = time.time()

        # build the starting set of possible_handles to be filtered
        possible_handles: Set[PrimaryObjectHandle]
        if id_list is not None:
            if tupleind is not None:
                # construct a dict from handle to corresponding tuple
                # this is used to efficiently transform final_list from a list of
                # handles to a list of tuples
                handle_tuple: Dict[PrimaryObjectHandle, Tuple] = {
                    data[tupleind]: data for data in cast(List[Tuple], id_list)
                }
                possible_handles = set(handle_tuple.keys())
            else:
                possible_handles = set(cast(List[PrimaryObjectHandle], id_list))
        elif tree:
            tree_handles = [handle for handle, obj in self.get_tree_cursor(db)]
            possible_handles = set(tree_handles)
        else:
            possible_handles = set(self.get_all_handles(db))

        res = self.apply_logical_op_to_all(db, possible_handles, apply_logical_op, user)

        # convert the filtered set of handles to the correct result type
        if id_list is not None and tupleind is not None:
            # convert the final_list of handles back to the corresponding final_list of tuples
            res = sorted(
                [handle_tuple[handle] for handle in res],
                key=lambda x: id_list.index(x),
            )
        elif tree:
            # sort final_list into the same order as traversed by get_tree_cursor
            res = sorted(res, key=lambda x: tree_handles.index(x))

        LOG.debug("Apply time: %s seconds", time.time() - start_time)

        for rule in self.flist:
            rule.requestreset()

        return res

    def get_object(self, db, handle):
        return db.get_person_from_handle(handle)


class GenericFamilyFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_all_handles(self, db):
        return db.get_family_handles()

    def get_cursor(self, db):
        return db.get_family_cursor()

    def make_obj(self):
        return Family()

    def find_from_handle(self, db, handle):
        return db.get_family_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_families()

    def get_object(self, db, handle):
        return db.get_family_from_handle(handle)


class GenericEventFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_all_handles(self, db):
        return db.get_event_handles()

    def get_cursor(self, db):
        return db.get_event_cursor()

    def make_obj(self):
        return Event()

    def find_from_handle(self, db, handle):
        return db.get_event_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_events()

    def get_object(self, db, handle):
        return db.get_event_from_handle(handle)


class GenericSourceFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_all_handles(self, db):
        return db.get_source_handles()

    def get_cursor(self, db):
        return db.get_source_cursor()

    def make_obj(self):
        return Source()

    def find_from_handle(self, db, handle):
        return db.get_source_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_sources()

    def get_object(self, db, handle):
        return db.get_source_from_handle(handle)


class GenericCitationFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_all_handles(self, db):
        return db.get_citation_handles()

    def get_cursor(self, db):
        return db.get_citation_cursor()

    def get_tree_cursor(self, db):
        return db.get_citation_cursor()

    def make_obj(self):
        return Citation()

    def find_from_handle(self, db, handle):
        return db.get_citation_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_citations()

    def get_object(self, db, handle):
        return db.get_citation_from_handle(handle)


class GenericPlaceFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_all_handles(self, db):
        return db.get_place_handles()

    def get_cursor(self, db):
        return db.get_place_cursor()

    def get_tree_cursor(self, db):
        return db.get_place_tree_cursor()

    def make_obj(self):
        return Place()

    def find_from_handle(self, db, handle):
        return db.get_place_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_places()

    def get_object(self, db, handle):
        return db.get_place_from_handle(handle)


class GenericMediaFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_all_handles(self, db):
        return db.get_media_handles()

    def get_cursor(self, db):
        return db.get_media_cursor()

    def make_obj(self):
        return Media()

    def find_from_handle(self, db, handle):
        return db.get_media_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_media()

    def get_object(self, db, handle):
        return db.get_media_from_handle(handle)


class GenericRepoFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_all_handles(self, db):
        return db.get_repository_handles()

    def get_cursor(self, db):
        return db.get_repository_cursor()

    def make_obj(self):
        return Repository()

    def find_from_handle(self, db, handle):
        return db.get_repository_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_repositories()

    def get_object(self, db, handle):
        return db.get_repository_from_handle(handle)


class GenericNoteFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_all_handles(self, db):
        return db.get_note_handles()

    def get_cursor(self, db):
        return db.get_note_cursor()

    def make_obj(self):
        return Note()

    def find_from_handle(self, db, handle):
        return db.get_note_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_notes()

    def get_object(self, db, handle):
        return db.get_note_from_handle(handle)


def GenericFilterFactory(namespace):
    if namespace == "Person":
        return GenericFilter
    elif namespace == "Family":
        return GenericFamilyFilter
    elif namespace == "Event":
        return GenericEventFilter
    elif namespace == "Source":
        return GenericSourceFilter
    elif namespace == "Citation":
        return GenericCitationFilter
    elif namespace == "Place":
        return GenericPlaceFilter
    elif namespace == "Media":
        return GenericMediaFilter
    elif namespace == "Repository":
        return GenericRepoFilter
    elif namespace == "Note":
        return GenericNoteFilter


class DeferredFilter(GenericFilter):
    """
    Filter class allowing for deferred translation of the filter name
    """

    def __init__(self, filter_name, person_name):
        GenericFilter.__init__(self, None)
        self.name_pair = [filter_name, person_name]

    def get_name(self, ulocale=glocale):
        """
        return the filter name, possibly translated

        If ulocale is passed in (a :class:`.GrampsLocale`) then
        the translated value will be returned instead.

        :param ulocale: allow deferred translation of strings
        :type ulocale: a :class:`.GrampsLocale` instance
        """
        self._ = ulocale.translation.gettext
        if self.name_pair[1]:
            return self._(self.name_pair[0]) % self.name_pair[1]
        return self._(self.name_pair[0])


class DeferredFamilyFilter(GenericFamilyFilter):
    """
    Filter class allowing for deferred translation of the filter name
    """

    def __init__(self, filter_name, family_name):
        GenericFamilyFilter.__init__(self, None)
        self.name_pair = [filter_name, family_name]

    def get_name(self, ulocale=glocale):
        """
        return the filter name, possibly translated

        If ulocale is passed in (a :class:`.GrampsLocale`) then
        the translated value will be returned instead.

        :param ulocale: allow deferred translation of strings
        :type ulocale: a :class:`.GrampsLocale` instance
        """
        self._ = ulocale.translation.gettext
        if self.name_pair[1]:
            return self._(self.name_pair[0]) % self.name_pair[1]
        return self._(self.name_pair[0])



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2010  Benny Malengier
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import MatchesSourceFilterBase


# -------------------------------------------------------------------------
#
# MatchesFilter
#
# -------------------------------------------------------------------------
class MatchesSourceFilter(MatchesSourceFilterBase):
    """
    Rule that checks against another filter.
    """

    labels = [_("Source filter name:")]
    name = _("Events with source matching the <source filter>")
    description = _(
        "Matches events with sources that match the " "specified source filter name"
    )
    category = _("Citation/source filters")

    # we want to have this filter show source filters
    namespace = "Source"



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Gary Burton
# Copyright (C) 2009       Nick Hall
# Copyright (C) 2010       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""
Person list View
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.plugins.lib.libpersonview import BasePersonView
from gramps.gui.views.treemodels.peoplemodel import PersonListModel

# -------------------------------------------------------------------------
#
# Internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# PersonListView
#
# -------------------------------------------------------------------------
class PersonListView(BasePersonView):
    """
    A hierarchical view of the top three levels of places.
    """

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        BasePersonView.__init__(
            self,
            pdata,
            dbstate,
            uistate,
            _("Person View"),
            PersonListModel,
            nav_group=nav_group,
        )

    def get_config_name(self):
        return __name__



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#
# Portuguese version translated by Duarte Loreto <happyguy_pt@hotmail.com>, 2007.
# Based on the Spanish file.

"""
Portuguese-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Portuguese parser
#
# -------------------------------------------------------------------------
class DateParserPT(DateParser):
    modifier_to_int = {
        "antes de": Date.MOD_BEFORE,
        "antes": Date.MOD_BEFORE,
        "ant.": Date.MOD_BEFORE,
        "ant": Date.MOD_BEFORE,
        "depois de": Date.MOD_AFTER,
        "depois": Date.MOD_AFTER,
        "dep.": Date.MOD_AFTER,
        "dep": Date.MOD_AFTER,
        "aprox.": Date.MOD_ABOUT,
        "aprox": Date.MOD_ABOUT,
        "apr.": Date.MOD_ABOUT,
        "apr": Date.MOD_ABOUT,
        "cerca de": Date.MOD_ABOUT,
        "ca.": Date.MOD_ABOUT,
        "ca": Date.MOD_ABOUT,
        "c.": Date.MOD_ABOUT,
        "por volta de": Date.MOD_ABOUT,
        "por volta": Date.MOD_ABOUT,
        "pvd.": Date.MOD_ABOUT,
        "de": Date.MOD_FROM,
        "até": Date.MOD_TO,
    }

    calendar_to_int = {
        "gregoriano": Date.CAL_GREGORIAN,
        "g": Date.CAL_GREGORIAN,
        "juliano": Date.CAL_JULIAN,
        "j": Date.CAL_JULIAN,
        "hebreu": Date.CAL_HEBREW,
        "h": Date.CAL_HEBREW,
        "islâmico": Date.CAL_ISLAMIC,
        "i": Date.CAL_ISLAMIC,
        "revolucionário": Date.CAL_FRENCH,
        "r": Date.CAL_FRENCH,
        "persa": Date.CAL_PERSIAN,
        "p": Date.CAL_PERSIAN,
        "swedish": Date.CAL_SWEDISH,
        "s": Date.CAL_SWEDISH,
    }

    quality_to_int = {
        "estimado": Date.QUAL_ESTIMATED,
        "estimada": Date.QUAL_ESTIMATED,
        "est.": Date.QUAL_ESTIMATED,
        "est": Date.QUAL_ESTIMATED,
        "calc.": Date.QUAL_CALCULATED,
        "calc": Date.QUAL_CALCULATED,
        "calculado": Date.QUAL_CALCULATED,
        "calculada": Date.QUAL_CALCULATED,
    }

    def init_strings(self):
        DateParser.init_strings(self)
        _span_1 = ["de"]
        _span_2 = ["a"]
        _range_1 = ["entre", r"ent\.", "ent"]
        _range_2 = ["e"]
        self._span = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_span_1), "|".join(_span_2)),
            re.IGNORECASE,
        )
        self._range = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_range_1), "|".join(_range_2)),
            re.IGNORECASE,
        )


# -------------------------------------------------------------------------
#
# Portuguese display
#
# -------------------------------------------------------------------------
class DateDisplayPT(DateDisplay):
    """
    Portuguese language date display class.
    """

    long_months = (
        "",
        "Janeiro",
        "Fevereiro",
        "Março",
        "Abril",
        "Maio",
        "Junho",
        "Julho",
        "Agosto",
        "Setembro",
        "Outubro",
        "Novembro",
        "Dezembro",
    )

    short_months = (
        "",
        "Jan",
        "Fev",
        "Mar",
        "Abr",
        "Mai",
        "Jun",
        "Jul",
        "Ago",
        "Set",
        "Out",
        "Nov",
        "Dez",
    )

    calendar = ("", "Juliano", "Hebreu", "Revolucionário", "Persa", "Islâmico", "Sueco")

    _mod_str = ("", "antes de ", "depois de ", "por volta de ", "", "", "")

    _qual_str = ("", "estimado ", "calculado ")

    formats = (
        "AAAA-MM-DD (ISO)",
        "Numérica",
        "Mês Dia, Ano",
        "MÊS Dia, Ano",
        "Dia Mês, Ano",
        "Dia MÊS, Ano",
    )
    # this must agree with DateDisplayEn's "formats" definition
    # (since no locale-specific _display_gregorian exists, here)

    def display(self, date):
        """
        Return a text string representing the date.
        """
        mod = date.get_modifier()
        cal = date.get_calendar()
        qual = date.get_quality()
        start = date.get_start_date()
        newyear = date.get_new_year()

        qual_str = self._qual_str[qual]

        if mod == Date.MOD_TEXTONLY:
            return date.get_text()
        elif start == Date.EMPTY:
            return ""
        elif mod == Date.MOD_SPAN:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "de", d1, "a", d2, scal)
        elif mod == Date.MOD_RANGE:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "entre", d1, "e", d2, scal)
        else:
            text = self.display_cal[date.get_calendar()](start)
            scal = self.format_extras(cal, newyear)
            return "%s%s%s%s" % (qual_str, self._mod_str[mod], text, scal)


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    (
        "pt_PT",
        "pt_PT.UTF-8",
        "pt_BR",
        "pt_BR.UTF-8",
        "pt" "portuguese",
        "Portuguese",
        ("%d-%m-%Y",),
    ),
    DateParserPT,
    DateDisplayPT,
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011 Helge Herz
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#
# gen.filters.rules/Source/_HasRepositoryCallNumberRef.py

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Source
from ....db import Database


_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
# "Sources which reference repositories by a special Call Name"
# -------------------------------------------------------------------------
class HasRepositoryCallNumberRef(Rule):
    """Sources which reference repositories by a special Call Number"""

    labels = [_("Text:")]
    name = _('Sources with repository reference containing <text> in "Call Number"')
    description = _(
        "Matches sources with a repository reference\n"
        'containing a substring in "Call Number"'
    )
    category = _("General filters")
    allow_regex = True

    def apply_to_one(self, db: Database, obj: Source) -> bool:
        for repo_ref in obj.reporef_list:
            if self.match_substring(0, repo_ref.call_number):
                return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2005  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""
Secondary Object class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from abc import abstractmethod

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .baseobj import BaseObject


# -------------------------------------------------------------------------
#
# SecondaryObject
#
# -------------------------------------------------------------------------
class SecondaryObject(BaseObject):
    """
    The SecondaryObject is the base class for all secondary objects in the
    database.
    """

    @abstractmethod
    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """

    @abstractmethod
    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """

    def is_equal(self, source):
        """
        Check if two secondary objects are equivalent.
        """
        return self.serialize() == source.serialize()

    def is_equivalent(self, other):
        """
        Return if this object is equivalent to other.

        Should be overwritten by objects that inherit from this class.
        """



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011  Jerome Rapinat
# Copyright (C) 2011  Douglas S. Blank
# Copyright (C) 2011  Benny Malengier
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#
# gen.filters.rules/_MatchesSourceConfidenceBase.py

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ...lib.citationbase import CitationBase
from ...db import Database


# -------------------------------------------------------------------------
# "Confidence level"
# Sources of an attribute of an event are ignored
# -------------------------------------------------------------------------
class MatchesSourceConfidenceBase(Rule):
    """Objects with a specific confidence level on 'direct' Source references"""

    labels = ["Confidence level:"]
    name = "Object with at least one direct source >= <confidence level>"
    description = (
        "Matches objects with at least one direct source with confidence level(s)"
    )
    category = _("Citation/source filters")

    def apply_to_one(self, db: Database, obj: CitationBase) -> bool:
        required_conf = int(self.list[0])
        for citation_handle in obj.citation_list:
            citation = db.get_citation_from_handle(citation_handle)
            if required_conf <= citation.confidence:
                return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
import re
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ...lib.notebase import NoteBase
from ...db import Database


# -------------------------------------------------------------------------
# Objects having notes that contain a substring or match a regular expression
# -------------------------------------------------------------------------
class HasNoteRegexBase(Rule):
    """Objects having notes containing <text>."""

    labels = [_("Text:")]
    name = "Objects having notes containing <text>"
    description = (
        "Matches objects whose notes contain a substring "
        "or match a regular expression"
    )
    category = _("General filters")
    allow_regex = True

    def apply_to_one(self, db: Database, obj: NoteBase) -> bool:
        for handle in obj.note_list:
            note = db.get_note_from_handle(handle)
            if self.match_substring(0, str(note.text)):
                return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2008       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""
Option class representing a boolean (yes/no, true/false).
"""

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from . import Option


# -------------------------------------------------------------------------
#
# BooleanOption class
#
# -------------------------------------------------------------------------
class BooleanOption(Option):
    """
    This class describes an option that is a boolean (True or False).
    """

    def __init__(self, label, value):
        """
        :param label: A friendly label to be applied to this option.
            Example: "Exclude living people"
        :type label: string
        :param value: An initial value for this option.
            Example: True
        :type value: boolean
        :return: nothing
        """
        Option.__init__(self, label, value)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""
Package providing filter rules for Gramps.
"""

from ._allmedia import AllMedia
from ._hasidof import HasIdOf
from ._regexpidof import RegExpIdOf
from ._hascitation import HasCitation
from ._hasnoteregexp import HasNoteRegexp
from ._hasnotematchingsubstringof import HasNoteMatchingSubstringOf
from ._hasreferencecountof import HasReferenceCountOf
from ._hassourcecount import HasSourceCount
from ._hassourceof import HasSourceOf
from ._mediaprivate import MediaPrivate
from ._matchesfilter import MatchesFilter
from ._matchessourceconfidence import MatchesSourceConfidence
from ._hasmedia import HasMedia
from ._hasattribute import HasAttribute
from ._changedsince import ChangedSince
from ._hastag import HasTag

editor_rule_list = [
    AllMedia,
    HasCitation,
    HasIdOf,
    RegExpIdOf,
    HasNoteRegexp,
    HasReferenceCountOf,
    HasSourceCount,
    HasSourceOf,
    MediaPrivate,
    MatchesFilter,
    MatchesSourceConfidence,
    HasMedia,
    HasAttribute,
    ChangedSince,
    HasTag,
]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2022 Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""
Utilities used in unit tests.
"""

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from ..utils.grampslocale import GrampsLocale

# -------------------------------------------------------------------------
#
# Unit test utilities
#
# -------------------------------------------------------------------------

parser = GrampsLocale(lang="en").date_parser
displayer = GrampsLocale().date_displayer


def localize_date(date_str):
    """
    Translate a date into the current locale.
    """
    date = parser.parse(date_str)
    return displayer.display(date)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ._hastextmatchingsubstringof import HasTextMatchingSubstringOf


# -------------------------------------------------------------------------
# "HasTextMatchingRegexOf"
# -------------------------------------------------------------------------
class HasTextMatchingRegexpOf(HasTextMatchingSubstringOf):
    """This is wrapping HasTextMatchingSubstringOf to enable the regex_match
    parameter.

    """

    def __init__(self, list, use_regex=False, use_case=False):
        HasTextMatchingSubstringOf.__init__(self, list, use_regex, use_case)

    def prepare(self, db, user):
        self.db = db
        self.person_map = set()
        self.event_map = set()
        self.source_map = set()
        self.repo_map = set()
        self.family_map = set()
        self.place_map = set()
        self.media_map = set()
        self.case_sensitive = False
        self.regexp_match = True
        self.cache_sources()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#               2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from copy import deepcopy

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.lib import NoteType
from gramps.gen.db import DbTxn
from gramps.gen.const import URL_MANUAL_SECT2

from .displaytabs import NoteTab, AddrEmbedList, WebEmbedList, SourceBackRefList
from ..widgets import MonitoredEntry, PrivacyButton, MonitoredDataType
from .editreference import RefTab, EditReference
from ..glade import Glade

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT2
WIKI_HELP_SEC = "Repository_Reference_Editor"


# -------------------------------------------------------------------------
#
# EditRepoRef class
#
# -------------------------------------------------------------------------
class EditRepoRef(EditReference):
    def __init__(self, state, uistate, track, source, source_ref, update):
        EditReference.__init__(self, state, uistate, track, source, source_ref, update)
        self.original = deepcopy(source.serialize())

    def _local_init(self):
        self.top = Glade()
        self.set_window(
            self.top.toplevel,
            self.top.get_object("repo_title"),
            _("Repository Reference Editor"),
        )
        self.setup_configs("interface.repo-ref", 600, 450)

        self.define_warn_box(self.top.get_object("warn_box"))
        self.define_expander(self.top.get_object("src_expander"))

        tblref = self.top.get_object("table70")
        notebook = self.top.get_object("notebook_ref")
        # recreate start page as GrampsTab
        notebook.remove_page(0)
        self.reftab = RefTab(
            self.dbstate, self.uistate, self.track, _("General"), tblref
        )
        tblref = self.top.get_object("table69")
        notebook = self.top.get_object("notebook_src")
        # recreate start page as GrampsTab
        notebook.remove_page(0)
        self.primtab = RefTab(
            self.dbstate, self.uistate, self.track, _("_General"), tblref
        )
        self.track_ref_for_deletion("primtab")

    def _connect_signals(self):
        self.define_ok_button(self.top.get_object("ok"), self.ok_clicked)
        self.define_cancel_button(self.top.get_object("cancel"))
        self.define_help_button(
            self.top.get_object("help"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("repository-rebuild", self.close)
        self._add_db_signal("repository-delete", self.check_for_close)

    def _setup_fields(self):
        self.callno = MonitoredEntry(
            self.top.get_object("call_number"),
            self.source_ref.set_call_number,
            self.source_ref.get_call_number,
            self.db.readonly,
        )

        self.gid = MonitoredEntry(
            self.top.get_object("gid"),
            self.source.set_gramps_id,
            self.source.get_gramps_id,
            self.db.readonly,
        )

        self.privacy = PrivacyButton(
            self.top.get_object("private"), self.source, self.db.readonly
        )

        self.ref_privacy = PrivacyButton(
            self.top.get_object("private_ref"), self.source_ref, self.db.readonly
        )

        self.title = MonitoredEntry(
            self.top.get_object("repo_name"),
            self.source.set_name,
            self.source.get_name,
            self.db.readonly,
        )

        self.type_selector = MonitoredDataType(
            self.top.get_object("media_type"),
            self.source_ref.set_media_type,
            self.source_ref.get_media_type,
            self.db.readonly,
            self.db.get_source_media_types(),
        )

        self.media_type_selector = MonitoredDataType(
            self.top.get_object("repo_type"),
            self.source.set_type,
            self.source.get_type,
            self.db.readonly,
            self.db.get_repository_types(),
        )

    def _create_tabbed_pages(self):
        """
        Create the notebook tabs and inserts them into the main
        window.
        """

        notebook_src = self.top.get_object("notebook_src")
        notebook_ref = self.top.get_object("notebook_ref")

        self._add_tab(notebook_src, self.primtab)
        self._add_tab(notebook_ref, self.reftab)

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_note_list(),
            "reporef_editor_shared_notes",
            notetype=NoteType.REPO,
        )
        self._add_tab(notebook_src, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self.comment_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.source_ref.get_note_list(),
            "reporef_editor_ref_notes",
            notetype=NoteType.REPOREF,
        )
        self._add_tab(notebook_ref, self.comment_tab)
        self.track_ref_for_deletion("comment_tab")

        self.address_tab = AddrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_address_list(),
            "reporef_editor_shared_address",
        )
        self._add_tab(notebook_src, self.address_tab)
        self.track_ref_for_deletion("address_tab")

        self.web_list = WebEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_url_list(),
            "reporef_editor_shared_internet",
        )
        self._add_tab(notebook_src, self.web_list)
        self.track_ref_for_deletion("web_list")

        self.backref_tab = SourceBackRefList(
            self.dbstate,
            self.uistate,
            self.track,
            self.db.find_backlink_handles(self.source.handle),
            "reporef_editor_shared_references",
            self.enable_warnbox,
        )
        self._add_tab(notebook_src, self.backref_tab)
        self.track_ref_for_deletion("backref_tab")

        self._setup_notebook_tabs(notebook_src)
        self._setup_notebook_tabs(notebook_ref)

    def build_menu_names(self, sourceref):
        if self.source:
            source_name = self.source.get_name()
            submenu_label = _("Repository: %s") % source_name
        else:
            submenu_label = _("New Repository")
        return (_("Repo Reference Editor"), submenu_label)

    def ok_clicked(self, obj):
        if self.source.handle:
            # only commit if it has changed
            if self.source.serialize() != self.original:
                with DbTxn(_("Modify Repository"), self.db) as trans:
                    self.db.commit_repository(self.source, trans)
        else:
            if self.check_for_duplicate_id("Repository"):
                return
            with DbTxn(_("Add Repository"), self.db) as trans:
                self.db.add_repository(self.source, trans)
            self.source_ref.ref = self.source.handle

        if self.update:
            self.update((self.source_ref, self.source))

        self.close()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""
LDS Ordinance class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from warnings import warn

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .citationbase import CitationBase
from .const import DIFFERENT, EQUAL, IDENTICAL
from .datebase import DateBase
from .notebase import NoteBase
from .placebase import PlaceBase
from .privacybase import PrivacyBase
from .secondaryobj import SecondaryObject

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# LdsOrd
#
# -------------------------------------------------------------------------
class LdsOrd(SecondaryObject, CitationBase, NoteBase, DateBase, PlaceBase, PrivacyBase):
    """
    Class that contains information about LDS Ordinances.

    LDS ordinances are similar to events, but have very specific additional
    information related to data collected by the Church of Jesus Christ
    of Latter-day Saints (Mormon church). The LDS church is the largest
    source of genealogical information in the United States.
    """

    BAPTISM = 0
    ENDOWMENT = 1
    SEAL_TO_PARENTS = 2
    SEAL_TO_SPOUSE = 3
    CONFIRMATION = 4
    INITIATORY = 5

    DEFAULT_TYPE = BAPTISM

    STATUS_NONE = 0
    STATUS_BIC = 1
    STATUS_CANCELED = 2
    STATUS_CHILD = 3
    STATUS_CLEARED = 4
    STATUS_COMPLETED = 5
    STATUS_DNS = 6
    STATUS_INFANT = 7
    STATUS_PRE_1970 = 8
    STATUS_QUALIFIED = 9
    STATUS_DNS_CAN = 10
    STATUS_STILLBORN = 11
    STATUS_SUBMITTED = 12
    STATUS_UNCLEARED = 13

    DEFAULT_STATUS = STATUS_NONE

    _TYPE_MAP = [
        (BAPTISM, _("Baptism"), "baptism"),
        (CONFIRMATION, _("Confirmation"), "confirmation"),
        (INITIATORY, _("Initiatory"), "initiatory"),
        (ENDOWMENT, _("Endowment"), "endowment"),
        (SEAL_TO_PARENTS, _("Sealed to Parents"), "sealed_to_parents"),
        (SEAL_TO_SPOUSE, _("Sealed to Spouse"), "sealed_to_spouse"),
    ]

    _STATUS_MAP = [
        (STATUS_NONE, _("<No Status>"), ""),
        (STATUS_BIC, _("Born in Covenant"), "BIC"),
        (STATUS_CANCELED, _("Canceled"), "Canceled"),
        (STATUS_CHILD, _("Child"), "Child"),
        (STATUS_CLEARED, _("Cleared"), "Cleared"),
        (STATUS_COMPLETED, _("Completed"), "Completed"),
        (STATUS_DNS, _("Do not seal"), "DNS"),
        (STATUS_INFANT, _("Infant"), "Infant"),
        (STATUS_PRE_1970, _("Pre-1970"), "Pre-1970"),
        (STATUS_QUALIFIED, _("Qualified"), "Qualified"),
        (STATUS_DNS_CAN, _("Do not seal/Cancel"), "DNS/CAN"),
        (STATUS_STILLBORN, _("Stillborn"), "Stillborn"),
        (STATUS_SUBMITTED, _("Submitted"), "Submitted"),
        (STATUS_UNCLEARED, _("Uncleared"), "Uncleared"),
    ]

    def __init__(self, source=None):
        """Create a LDS Ordinance instance."""
        CitationBase.__init__(self, source)
        NoteBase.__init__(self, source)
        DateBase.__init__(self, source)
        PlaceBase.__init__(self, source)
        PrivacyBase.__init__(self, source)

        if source:
            self.type = source.type
            self.famc = source.famc
            self.temple = source.temple
            self.status = source.status
        else:
            self.type = LdsOrd.DEFAULT_TYPE
            self.famc = None
            self.temple = ""
            self.status = LdsOrd.DEFAULT_STATUS

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (
            CitationBase.serialize(self),
            NoteBase.serialize(self),
            DateBase.serialize(self),
            self.type,
            self.place,
            self.famc,
            self.temple,
            self.status,
            self.private,
        )

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        (
            citation_list,
            note_list,
            date,
            self.type,
            self.place,
            self.famc,
            self.temple,
            self.status,
            self.private,
        ) = data
        CitationBase.unserialize(self, citation_list)
        NoteBase.unserialize(self, note_list)
        DateBase.unserialize(self, date)
        return self

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        # pylint: disable=import-outside-toplevel
        from .date import Date

        return {
            "type": "object",
            "title": _("LDS Ordinance"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "citation_list": {
                    "type": "array",
                    "title": _("Citations"),
                    "items": {"type": "string", "maxLength": 50},
                },
                "note_list": {
                    "type": "array",
                    "title": _("Notes"),
                    "items": {"type": "string", "maxLength": 50},
                },
                "date": Date.get_schema(),
                "type": {"type": "integer", "title": _("Type")},
                "place": {"type": "string", "title": _("Place")},
                "famc": {"type": ["null", "string"], "title": _("Family")},
                "temple": {"type": "string", "title": _("Temple")},
                "status": {"type": "integer", "title": _("Status")},
                "private": {"type": "boolean", "title": _("Private")},
            },
        }

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.temple]
        # return [self.temple,self.get_date()]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return []

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return []

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        ret = (
            self.get_referenced_note_handles() + self.get_referenced_citation_handles()
        )
        if self.place:
            ret += [("Place", self.place)]
        if self.famc:
            ret += [("Family", self.famc)]
        return ret

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return []

    def is_equivalent(self, other):
        """
        Return if this ldsord is equivalent, that is agrees in date, temple,
        place, status, sealed_to, to other.

        :param other: The ldsord to compare this one to.
        :type other: LdsOrd
        :returns: Constant indicating degree of equivalence.
        :rtype: int
        """
        if (
            self.type != other.type
            or self.get_date_object() != other.get_date_object()
            or self.temple != other.temple
            or self.status != other.status
            or self.famc != other.famc
        ):
            return DIFFERENT
        if self.is_equal(other):
            return IDENTICAL
        return EQUAL

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this ldsord.

        Lost: type, date, temple, place, status, sealed_to of acquistion.

        :param acquisition: The ldsord to merge with the present ldsord.
        :type acquisition: LdsOrd
        """
        self._merge_privacy(acquisition)
        self._merge_note_list(acquisition)
        self._merge_citation_list(acquisition)

    def get_type(self):
        """
        Return the type of the Event.

        :returns: Type of the Event
        :rtype: tuple
        """
        return self.type

    def set_type(self, ord_type):
        """
        Set the type of the LdsOrd to the passed (int,str) tuple.

        :param ord_type: Type to assign to the LdsOrd
        :type ord_type: tuple
        """
        self.type = ord_type

    def set_family_handle(self, family):
        """Set the Family database handle associated with the LDS ordinance."""
        self.famc = family

    def get_family_handle(self):
        """Get the Family database handle associated with the LDS ordinance."""
        return self.famc

    def set_status(self, val):
        """
        Set the status of the LDS ordinance.

        The status is a text string that matches a predefined set of strings.
        """
        self.status = val

    def get_status(self):
        """Get the status of the LDS ordinance."""
        return self.status

    def set_temple(self, temple):
        """Set the temple associated with the ordinance."""
        self.temple = temple

    def get_temple(self):
        """Get the temple associated with the ordinance."""
        return self.temple

    def is_empty(self):
        """Return 1 if the ordinance is actually empty."""
        if (
            self.famc
            or (self.date and not self.date.is_empty())
            or self.temple
            or self.status
            or self.place
        ):
            return False
        return True

    def are_equal(self, other):
        """Return 1 if the specified ordinance is the same as the instance."""
        warn("Use is_equal instead are_equal", DeprecationWarning, 2)
        return self.is_equal(other)

    def type2xml(self):
        """
        Return type-representing string suitable for XML.
        """
        for item in LdsOrd._TYPE_MAP:
            if item[0] == self.type:
                return item[2]
        return ""

    def type2str(self):
        """
        Return type-representing string suitable for UI (translated).
        """
        for item in LdsOrd._TYPE_MAP:
            if item[0] == self.type:
                return item[1]
        return ""

    def set_type_from_xml(self, xml_str):
        """
        Set type based on a given string from XML.

        Return boolean on success.
        """
        for item in LdsOrd._TYPE_MAP:
            if item[2] == xml_str:
                self.type = item[0]
                return True
        return False

    def status2xml(self):
        """
        Return status-representing string suitable for XML.
        """
        for item in LdsOrd._STATUS_MAP:
            if item[0] == self.status:
                return item[2]
        return ""

    def status2str(self):
        """
        Return status-representing string suitable for UI (translated).
        """
        for item in LdsOrd._STATUS_MAP:
            if item[0] == self.status:
                return item[1]
        return ""

    def set_status_from_xml(self, xml_str):
        """
        Set status based on a given string from XML.

        Return boolean on success.
        """
        for item in LdsOrd._STATUS_MAP:
            if item[2] == xml_str:
                self.status = item[0]
                return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Gary Burton
# Copyright (C) 2019  Matthias Kemmer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .._hasattributebase import HasAttributeBase


# -------------------------------------------------------------------------
#
# HasAttribute
#
# -------------------------------------------------------------------------
class HasAttribute(HasAttributeBase):
    """Rule that checks for a citation with a particular attribute"""

    labels = [_("Citation attribute:"), _("Value:")]
    name = _("Citations with the attribute <attribute>")
    description = _("Matches citations with the attribute " "of a particular value")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Set
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# IsLessThanNthGenerationDescendantOf
#
# -------------------------------------------------------------------------
class IsLessThanNthGenerationDescendantOf(Rule):
    """Rule that checks for a person that is a descendant of a specified person
    not more than N generations away"""

    labels = [_("ID:"), _("Number of generations:")]
    name = _("Descendants of <person> not more than " "<N> generations away")
    category = _("Descendant filters")
    description = _(
        "Matches people that are descendants of a "
        "specified person not more than N generations away"
    )

    def prepare(self, db: Database, user):
        self.db = db
        self.selected_handles: Set[str] = set()
        try:
            root_person = db.get_person_from_gramps_id(self.list[0])
            self.init_list(root_person, 0)
        except:
            pass

    def reset(self):
        self.selected_handles.clear()

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.handle in self.selected_handles

    def init_list(self, person: Person | None, gen: int):
        if not person or person.handle in self.selected_handles:
            # if we have been here before, skip
            return
        if gen:
            self.selected_handles.add(person.handle)
            if gen >= int(self.list[1]):
                return

        for fam_id in person.family_list:
            fam = self.db.get_family_from_handle(fam_id)
            if fam:
                for child_ref in fam.child_ref_list:
                    self.init_list(
                        self.db.get_person_from_handle(child_ref.ref), gen + 1
                    )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008, 2012  Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#
# gen/proxy/__init__.py

__all__ = ["filter", "living", "private", "proxybase", "referencedbyselection"]

from .filter import FilterProxyDb
from .living import LivingProxyDb
from .private import PrivateProxyDb
from .referencedbyselection import ReferencedBySelectionProxyDb
from .cache import CacheProxyDb



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2010       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import time

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.filters import reload_custom_filters

_ = glocale.translation.gettext
from bisect import insort_left
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

from ... import widgets
from ...dbguielement import DbGUIElement
from gramps.gen.config import config
from ...utils import no_match_primary_mask

from ...editors import EditFilter
import gramps.gen.filters

# import gramps.gen.filters.rules.place

_RETURN = Gdk.keyval_from_name("Return")
_KP_ENTER = Gdk.keyval_from_name("KP_Enter")


class SidebarFilter(DbGUIElement):
    _FILTER_WIDTH = -1
    _FILTER_ELLIPSIZE = Pango.EllipsizeMode.END

    def __init__(self, dbstate, uistate, namespace):
        self.signal_map = {
            "tag-add": self._tag_add,
            "tag-delete": self._tag_delete,
            "tag-update": self._tag_update,
            "tag-rebuild": self._tag_rebuild,
        }
        DbGUIElement.__init__(self, dbstate.db)

        self.position = 1
        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.grid = Gtk.Grid()
        self.vbox.pack_start(self.grid, False, False, 0)
        self.grid.set_border_width(6)
        self.grid.set_row_spacing(6)
        self.grid.set_column_spacing(6)
        self.apply_btn = Gtk.Button.new_with_mnemonic(_("_Find"))
        self.apply_btn.set_tooltip_text(
            _("This updates the view with the current filter parameters.")
        )
        self.clear_btn = Gtk.Button()
        self.clear_btn.set_tooltip_text(
            _(
                "This resets the filter parameters to empty state.  The 'Find' "
                "button should be used to actually update the view to its "
                "defaults."
            )
        )
        self.define_filter_btn = Gtk.Button(_("Define filter"))
        self.define_filter_btn.set_tooltip_text(
            _(
                "This opens a dialog to add a new custom filter "
                "based on the values given."
            )
        )

        self.msg_label = Gtk.Label()
        self.msg_label.set_halign(Gtk.Align.START)
        self.msg_label.set_margin_left(12)

        self._init_interface()
        uistate.connect("filters-changed", self.on_filters_changed)
        dbstate.connect("database-changed", self._db_changed)
        self.uistate = uistate
        self.dbstate = dbstate
        self.namespace = namespace
        self.__tag_list = []
        self._tag_rebuild()

    def _init_interface(self):
        self.create_widget()

        self.apply_btn.connect("clicked", self.clicked)

        hbox = Gtk.Box()
        hbox.show()
        image = Gtk.Image()
        image.set_from_icon_name("edit-undo", Gtk.IconSize.BUTTON)
        image.show()
        label = Gtk.Label(label=_("Reset"))
        label.show()
        hbox.pack_start(image, False, False, 0)
        hbox.pack_start(label, False, True, 0)
        hbox.set_spacing(4)

        self.clear_btn.add(hbox)
        self.clear_btn.connect("clicked", self.clear)
        self.clear_btn.set_halign(Gtk.Align.END)

        hbox = Gtk.ButtonBox()
        hbox.set_layout(Gtk.ButtonBoxStyle.START)
        hbox.set_spacing(6)
        hbox.set_border_width(12)
        hbox.add(self.apply_btn)
        hbox.add(self.clear_btn)
        hbox.show()
        self.vbox.pack_start(hbox, False, False, 0)

        self.define_filter_btn.connect("clicked", self.define_filter)
        hbox = Gtk.ButtonBox()
        hbox.set_layout(Gtk.ButtonBoxStyle.START)
        hbox.set_spacing(6)
        hbox.set_border_width(12)
        hbox.add(self.define_filter_btn)

        self.vbox.pack_start(hbox, False, False, 0)
        self.vbox.pack_start(self.msg_label, False, False, 0)

        self.vbox.show()

    def get_widget(self):
        return self.vbox

    def create_widget(self):
        pass

    def clear(self, obj):
        pass

    def clicked(self, obj):
        if not self.filter_is_ok():
            return
        self.uistate.set_busy_cursor(True)
        t1 = time.perf_counter()
        self.clicked_func()
        t2 = time.perf_counter()
        msg = _("Elapsed time: %.2fs") % (t2 - t1)
        self.msg_label.set_text(msg)
        self.msg_label.get_style_context().remove_class("error")
        self.uistate.set_busy_cursor(False)

    def clicked_func(self):
        pass

    def get_filter(self):
        pass

    def add_regex_entry(self, widget):
        hbox = Gtk.Box()
        hbox.pack_start(widget, False, False, 12)
        widget.connect("toggled", self.regex_selection)
        self.vbox.pack_start(hbox, False, False, 0)

    def add_regex_case(self, widget):
        hbox = Gtk.Box()
        hbox.pack_start(widget, False, False, 12)
        self.vbox.pack_start(hbox, False, False, 0)
        self.regex_selection()

    def regex_selection(self, widget=None):
        if self.sensitive_regex:
            if widget and widget.get_active():
                self.sensitive_regex.set_sensitive(True)
            else:
                self.sensitive_regex.set_active(False)
                self.sensitive_regex.set_sensitive(False)

    def add_text_entry(self, name, widget, tooltip=None):
        self.add_entry(name, widget)
        if isinstance(widget, widgets.DateEntry):
            widget.entry.connect("key-press-event", self.key_press)
        else:
            widget.connect("key-press-event", self.key_press)
        if tooltip:
            widget.set_tooltip_text(tooltip)

    def key_press(self, obj, event):
        if no_match_primary_mask(event.get_state()):
            if event.keyval in (_RETURN, _KP_ENTER):
                self.clicked(obj)
        return False

    def add_heading(self, heading):
        label = Gtk.Label()
        label.set_text("<b>%s</b>" % heading)
        label.set_use_markup(True)
        label.set_halign(Gtk.Align.START)
        self.grid.attach(label, 1, self.position, 1, 1)
        self.position += 1

    def add_entry(self, name, widget):
        if name:
            self.grid.attach(widgets.BasicLabel(name), 1, self.position, 1, 1)
        widget.set_hexpand(True)
        self.grid.attach(widget, 2, self.position, 2, 1)
        self.position += 1

    def on_filters_changed(self, namespace):
        """
        Called when filters are changed.
        """
        pass

    def _db_changed(self, db):
        """
        Called when the database is changed.
        """
        self._change_db(db)
        self.on_db_changed(db)
        self._tag_rebuild()

    def on_db_changed(self, db):
        """
        Called when the database is changed.
        """
        pass

    def _connect_db_signals(self):
        """
        Connect database signals defined in the signal map.
        """
        for sig in self.signal_map:
            self.callman.add_db_signal(sig, self.signal_map[sig])

    def _tag_add(self, handle_list):
        """
        Called when tags are added.
        """
        for handle in handle_list:
            tag = self.dbstate.db.get_tag_from_handle(handle)
            insort_left(self.__tag_list, (tag.get_name(), handle))
        self.on_tags_changed([item[0] for item in self.__tag_list])

    def _tag_update(self, handle_list):
        """
        Called when tags are updated.
        """
        for handle in handle_list:
            item = [item for item in self.__tag_list if item[1] == handle][0]
            self.__tag_list.remove(item)
            tag = self.dbstate.db.get_tag_from_handle(handle)
            insort_left(self.__tag_list, (tag.get_name(), handle))
        self.on_tags_changed([item[0] for item in self.__tag_list])

    def _tag_delete(self, handle_list):
        """
        Called when tags are deleted.
        """
        self.__tag_list = [
            item for item in self.__tag_list if item[1] not in handle_list
        ]
        self.on_tags_changed([item[0] for item in self.__tag_list])

    def _tag_rebuild(self):
        """
        Called when the tag list needs to be rebuilt.
        """
        self.__tag_list = []
        if self.dbstate.is_open():
            for handle in self.dbstate.db.get_tag_handles(sort_handles=True):
                tag = self.dbstate.db.get_tag_from_handle(handle)
                self.__tag_list.append((tag.get_name(), handle))
        self.on_tags_changed([item[0] for item in self.__tag_list])

    def on_tags_changed(self, tag_list):
        """
        Called when tags are changed.
        """
        pass

    def add_filter_entry(self, text, widget):
        """
        Adds the text and widget to GUI, with an Edit button.
        """
        hbox = Gtk.Box()
        hbox.pack_start(widget, True, True, 0)
        hbox.pack_start(
            widgets.SimpleButton("gtk-edit", self.edit_filter), False, False, 0
        )
        self.add_entry(text, hbox)

    def edit_filter(self, obj):
        """
        Callback which invokes the EditFilter dialog. Will create new
        filter if called if none is selected.
        """
        from ...editors import EditFilter
        from gramps.gen.filters import FilterList, GenericFilterFactory
        from gramps.gen.const import CUSTOM_FILTERS

        the_filter = None
        filterdb = FilterList(CUSTOM_FILTERS)
        filterdb.load()
        if self.generic.get_active() != 0:
            model = self.generic.get_model()
            node = self.generic.get_active_iter()
            if node:
                sel_filter = model.get_value(node, 1)
                # the_filter needs to be a particular object for editor
                for filt in filterdb.get_filters(self.namespace):
                    if filt.get_name() == sel_filter.get_name():
                        the_filter = filt
        else:
            the_filter = GenericFilterFactory(self.namespace)()
        if the_filter:
            EditFilter(
                self.namespace,
                self.dbstate,
                self.uistate,
                [],
                the_filter,
                filterdb,
                selection_callback=self.edit_filter_save,
            )

    def edit_filter_save(self, filterdb, filter_name):
        """
        If a filter changed, save them all. Reloads, and sets name.
        Takes the filter database, and the filter name edited.
        """
        from gramps.gen.filters import reload_custom_filters

        filterdb.save()
        reload_custom_filters()
        self.uistate.emit("filters-changed", (self.namespace,))
        self.set_filters_to_name(filter_name)

    def set_filters_to_name(self, filter_name):
        """
        Resets the Filter combobox to the edited/saved filter.
        """
        liststore = self.generic.get_model()
        iter = liststore.get_iter_first()
        while iter:
            filter = liststore.get_value(iter, 1)
            if filter and filter.name == filter_name:
                self.generic.set_active_iter(iter)
                break
            iter = liststore.iter_next(iter)

    def filter_is_ok(self):
        the_filter = self.get_filter()
        if the_filter is None:
            return True
        for rule in the_filter.get_rules():
            if isinstance(rule, gramps.gen.filters.rules.place.WithinArea):
                if rule.list[0] is None:  # No active place
                    msg = _("You should select a place when using the 'Within' rule")
                    self.msg_label.set_text(msg)
                    self.msg_label.get_style_context().add_class("error")
                    return False
        return True

    def define_filter(self, _obj):
        self.filterdb = gramps.gen.filters.CustomFilters
        the_filter = self.get_filter()
        if the_filter is None:
            self.msg_label.set_text(_("Supply at least one value"))
            self.msg_label.get_style_context().add_class("error")
            return

        if not self.filter_is_ok():
            return
        # fix some rules:
        new_rules = []
        for rule in the_filter.get_rules():
            # The Place rule WithinArea might have numeric values while custom_filters.xml only accepts strings.
            if isinstance(rule, gramps.gen.filters.rules.place.WithinArea):
                rule.list[1] = str(rule.list[1])
                rule.list[2] = str(rule.list[2])
            new_rules.append(rule)
        the_filter.flist = new_rules
        comment = _("Created by Filter gramplet on {today}").format(
            today=time.strftime("%Y-%m-%d", time.localtime())
        )
        the_filter.set_comment(comment)

        self.msg_label.set_text("")
        track = []
        EditFilter(
            self.namespace,
            self.dbstate,
            self.uistate,
            track,
            the_filter,
            self.filterdb,
            update=self.update,
        )

    def update(self):
        self.filterdb.save()
        reload_custom_filters()
        self.uistate.emit("filters-changed", (self.namespace,))



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2024  Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2025  Steve Youngs <steve@youngs.cc>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#
# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
import logging

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import List, Set, TYPE_CHECKING, Tuple
from .rules import Rule

if TYPE_CHECKING:
    from ._genericfilter import GenericFilter
from ..types import PrimaryObjectHandle

LOG = logging.getLogger(".filter.optimizer")


class Optimizer:
    def __init__(self, top_level_filter):
        self.top_level_filter = top_level_filter

    def compute_potential_handles_for_filter(
        self, filter: GenericFilter
    ) -> Tuple[Set[PrimaryObjectHandle] | None, Set[PrimaryObjectHandle] | None]:
        if len(filter.flist) == 0:
            return (None, None)

        handles_in = None
        handles_out = None
        for rule in filter.flist:
            if filter.logical_op == "and" or len(filter.flist) == 1:
                rule_in, rule_out = self.compute_potential_handles_for_rule(rule)
                if rule_in is not None:
                    if handles_in is None:
                        handles_in = rule_in
                    else:
                        handles_in = handles_in.intersection(rule_in)
                if rule_out is not None:
                    if handles_out is None:
                        handles_out = rule_out
                    else:
                        handles_out = handles_out.union(rule_out)

        if filter.invert:
            handles_in, handles_out = handles_out, handles_in

        return handles_in, handles_out

    def compute_potential_handles_for_rule(
        self,
        rule: Rule,
    ) -> Tuple[Set[PrimaryObjectHandle] | None, Set[PrimaryObjectHandle] | None]:
        """
        Find the the handles for a particular rule.
        """
        # Has to be of the appropriate type
        if hasattr(rule, "selected_handles"):
            return (rule.selected_handles, None)
        if hasattr(rule, "find_filter"):
            filter = rule.find_filter()
            if filter and self.is_same_namespace(filter):
                return self.compute_potential_handles_for_filter(filter)
        return (None, None)

    def is_same_namespace(self, filter):
        """
        Determine if the given filter is in the 'same namespace' as the top-level filter.

        In this context, 'same namespace' means that both filters are instances of the same class,
        as determined by comparing their class names.

        Parameters:
            filter: The filter object to compare against the top-level filter.

        Returns:
            True if both filters have the object type, False otherwise.
        """
        return type(self.top_level_filter.make_obj()) == type(filter.make_obj())



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010 Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "6.0"

# ------------------------------------------------------------------------
#
# Register default sidebars
#
# ------------------------------------------------------------------------

register(
    SIDEBAR,
    id="categorysidebar",
    name=_("Category Sidebar"),
    description=_("A sidebar to allow the selection of view categories"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="categorysidebar.py",
    authors=["Nick Hall"],
    authors_email=["nick__hall@hotmail.com"],
    sidebarclass="CategorySidebar",
    menu_label=_("Category"),
    order=START,
)

register(
    SIDEBAR,
    id="dropdownsidebar",
    name=_("Drop-down Sidebar"),
    description=_("Selection of categories and views from drop-down lists"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="dropdownsidebar.py",
    authors=["Nick Hall"],
    authors_email=["nick__hall@hotmail.com"],
    sidebarclass="DropdownSidebar",
    menu_label=_("Drop-Down"),
    order=END,
)

register(
    SIDEBAR,
    id="expandersidebar",
    name=_("Expander Sidebar"),
    description=_("Selection of views from lists with expanders"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="expandersidebar.py",
    authors=["Nick Hall"],
    authors_email=["nick__hall@hotmail.com"],
    sidebarclass="ExpanderSidebar",
    menu_label=_("Expander"),
    order=END,
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

__all__ = ["ExpandCollapseArrow"]

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

import logging

_LOG = logging.getLogger(".widgets.expandcollapsearrow")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk
from gramps.gen.constfunc import has_display

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
if has_display():
    HAND_CURSOR = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.HAND2
    )


# -------------------------------------------------------------------------
#
# Module functions
#
# -------------------------------------------------------------------------
def realize_cb(widget):
    widget.get_window().set_cursor(HAND_CURSOR)


# -------------------------------------------------------------------------
#
# ExpandCollapseArrow class
#
# -------------------------------------------------------------------------
class ExpandCollapseArrow(Gtk.EventBox):
    """
    Arrow to be used for expand/collapse of sections.

    .. note:: shadow does not work, we indicate action with realize_cb
    """

    def __init__(self, collapsed, onbuttonpress, pair):
        """
        Constructor for the ExpandCollapseArrow class.

        :param collapsed: True if arrow must be shown collapsed,
                          False otherwise
        :type collapsed: bool
        :param onbuttonpress: The callback function for button press
        :type onbuttonpress: callback
        :param pair: user param for onbuttonpress function
        """
        Gtk.EventBox.__init__(self)
        if collapsed:
            self.arrow = Gtk.Image.new_from_icon_name(
                "pan-end-symbolic", Gtk.IconSize.BUTTON
            )
            self.set_tooltip_text(_("Expand this section"))
        else:
            self.arrow = Gtk.Image.new_from_icon_name(
                "pan-down-symbolic", Gtk.IconSize.BUTTON
            )
            self.set_tooltip_text(_("Collapse this section"))
        self.add(self.arrow)
        self.connect("button-press-event", onbuttonpress, pair)
        self.connect("realize", realize_cb)
